from __future__ import annotations

import functools
from typing import Any, Union, List
from pysqream.column_buffer import ColumnsBuffer
from pysqream.globals import ROWS_PER_FLUSH, FETCH_MANY_DEFAULT, TYPE_MAPPER, BYTES_PER_FLUSH_LIMIT, TEXT_ITEM_SIZE
from pysqream.utils import (NotSupportedError, ProgrammingError, Error, ArraysAreDisabled, mark_dbapi_attribute, dbapi_method, validate_disk_space,
                            validate_file_size, get_unique_filename)
from pysqream.server.sqclient import SQClient
from pysqream.server.connection_params import ConnectionParams
from pysqream.logger import ContextLogger


class Cursor:
    """
    Represent a database cursor, which is used to manage the context of
    a statement (execute and fetch).

    Cursors created from the same connection are not
    isolated, i.e., any changes done to the database by a cursor are
    immediately visible by the other cursors.

    Base PEP 249 – Python Database API Specification v2.0
    https://peps.python.org/pep-0249/#id12
    """

    def __init__(self, conn_params: ConnectionParams, client: SQClient, connection_id: int, allow_array: bool, logger: ContextLogger):
        self.__logger = logger

        # DB-API must have attributes
        self.arraysize = mark_dbapi_attribute(FETCH_MANY_DEFAULT)
        self.rowcount = mark_dbapi_attribute(-1)
        self.lastrowid = mark_dbapi_attribute(None)
        self.description = mark_dbapi_attribute(None)

        # SQreamDB connection attributes
        self.__conn_params: ConnectionParams = conn_params
        self.__client: SQClient = client
        self.__connection_id: int = connection_id
        self.__allow_array: bool = allow_array

        # State management
        self.__open_statement: bool = False
        self.__is_cursor_closed: bool = False
        self.__is_connection_initiated_close: bool = False

        # Buffer and data management
        self.__buffer: ColumnsBuffer = ColumnsBuffer(self.__logger)  # flushing buffer every BUFFER_SIZE bytes
        self.__stmt_id: Union[id, None] = None  # For error handling when called out of order
        self.__statement_type: Union[str, None] = None
        self.__more_to_fetch: bool = False
        self.__capacity: int = 0

        # Data tracking
        self.__parsed_rows: List[Any] = []
        self.__row_size: int = 0
        self.__rows_per_flush: int = 0
        self.__cols: List = []
        self.__col_types: List = []
        self.__col_sizes: List = []
        self.__col_scales: List = []
        self.__unsorted_data_columns: List = []

    def __enter__(self) -> Cursor:
        """
        Implementation for context manager ("with" clause)
        """

        return self

    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:
        """
        Implementation for context manager ("with" clause)
        """

        self.close()

    def __iter__(self):
        """
        Implementation for iterating connection in for-in clause
        """

        for item in self.fetchall():
            yield item

    # DB-API Must have methods
    # ------------------------
    @dbapi_method
    def execute(self, statement: str, params: list[tuple[Any]] | tuple[[tuple[Any]]] | None = None,
                data_as: str = 'rows', amount: int | None = None) -> Cursor:
        """
        Execute a statement. If params was provided - compile statement first
        and replace all question marks on passed parameters.

        :param statement: str a statement to execute with placeholders
        :param params: list[tuple[Any]] | tuple[tuple[Any]]: sequence of parameters to ingest into query
        :param data_as: string type of passed params. Possible values (`alchemy_flat_list`, `numpy`, `rows` - default)
        :param amount: int - amount of values of given params to insert

        Examples:
            query: SELECT * FROM <table_name> WHERE id IN (?, ?, ?) AND price >= ?;
            params: [(1, 2, 3, 450.69),]

            query: SELECT * FROM <table_name> WHERE id = %s AND price < %s;
            params: [(1, 200.00),]

            query: INSERT INTO <table_name> (id, name, description, price) VALUES (?, ?, ?, ?);
            params: [(1, 'Ryan Gosling', 'Actor', 0.0),]
            for next params it is better to use `executemany`
            params: [(1, 'Ryan Gosling', 'Actor', 0.0), (2, 'Mark Wahlberg', 'No pain no gain', 150.0)]

            query: UPDATE <table_name> SET price = ?, description = ? WHERE name = ?;
            params: [(999.999, 'Best actor', 'Ryan Gosling'),]

            query: DELETE FROM <table_name> WHERE id = ? AND other like ?;
            params: [(404, 200),]
        """

        try:
            self.__verify_cur_open()
            self.__execute_sqream_statement(statement, params=params, data_as=data_as, amount=amount)
            self.__fill_description()
        except Exception as e:
            self.__close_stmt()
            raise e

        return self

    @dbapi_method
    def executemany(self, statement: str, rows_or_cols: list[tuple[Any]] | tuple[[tuple[Any]]] | None = None,
                    data_as: str = 'rows', amount: int | None = None) -> Cursor:
        """
        Execute a statement, preferably used for insertion

        :param statement: str a statement to execute with placeholders
        :param rows_or_cols: list[tuple[Any]] | tuple[tuple[Any]]: sequence of parameters to ingest into query
        :param data_as: string type of passed params. Possible values (`alchemy_flat_list`, `numpy`, `rows` - default)
        :param amount: int - amount of values of given params to insert
        """

        return self.execute(statement, params=rows_or_cols, data_as=data_as, amount=amount)

    @dbapi_method
    def fetchmany(self, size: int = None, data_as: str = 'rows', fetchone: bool = False):
        """
        Fetch an amount of result rows
        """

        self.__logger.info(f"Fetching {'all' if size == -1 else size} rows", connection_id=self.__connection_id, statement_id=self.__stmt_id)
        size = size or self.arraysize

        if self.__statement_type not in (None, 'SELECT'):
            self.__logger.log_and_raise(ProgrammingError, 'No open statement while attempting fetch operation',
                                        connection_id=self.__connection_id, statement_id=self.__stmt_id)

        if self.__more_to_fetch:
            if self.__prepare_params.get("local_file_write", False):
                self.__get_file_from_server()
            else:
                self.__fetch_and_parse(size, data_as)
                self.__logger.debug(f"Fetched {len(self.__parsed_rows) if self.__parsed_rows else 0} rows",
                                    connection_id=self.__connection_id, statement_id=self.__stmt_id)

        res = self.__parsed_rows[0:size if size != -1 else None]
        del self.__parsed_rows[:size if size != -1 else None]

        return (res if res else []) if not fetchone else (res[0] if res else None)

    @dbapi_method
    def fetchone(self, data_as: str = 'rows'):
        """
        Fetch one result row
        """

        if data_as not in ('rows',):
            self.__logger.log_and_raise(ProgrammingError, "Bad argument to fetchone()", connection_id=self.__connection_id, statement_id=self.__stmt_id)

        return self.fetchmany(1, data_as, fetchone=True)

    @dbapi_method
    def fetchall(self, data_as: str = 'rows'):
        """
        Fetch all result rows
        """

        if data_as not in ('rows',):
            self.__logger.log_and_raise(ProgrammingError, "Bad argument to fetchall()", connection_id=self.__connection_id, statement_id=self.__stmt_id)

        return self.fetchmany(-1, data_as)

    @dbapi_method
    def nextset(self):
        """
        No multiple result sets so currently always returns None
        """

        self.__logger.log_and_raise(NotSupportedError, "Nextset is not supported", connection_id=self.__connection_id)

    @dbapi_method
    def setinputsizes(self, sizes):
        self.__logger.log_and_raise(NotSupportedError, "Setinputsizes is not supported", connection_id=self.__connection_id)

    @dbapi_method
    def setoutputsize(self, size, column=None):
        self.__logger.log_and_raise(NotSupportedError, "Setoutputsize is not supported", connection_id=self.__connection_id)

    @dbapi_method
    def close(self):
        """
        Closes the statement against sqream, cleans the buffer and change the cursor status to closed
        """

        self.__close_stmt()
        self.__is_cursor_closed = True

    # Internal Methods
    # ----------------
    def get_statement_type(self) -> str:
        return self.__statement_type

    def get_statement_id(self) -> int:
        return self.__stmt_id

    def get_connection_id(self) -> int:
        return self.__connection_id

    def get_is_cursor_closed(self) -> bool:
        return self.__is_cursor_closed

    def __verify_cur_open(self) -> None:
        if self.__is_cursor_closed:
            self.__logger.log_and_raise(ProgrammingError,
                                        f"{'Connection' if self.__is_connection_initiated_close else 'Cursor'}"
                                        f" has been closed", connection_id=self.__connection_id)

    def __execute_sqream_statement(self, statement: str, params: list[Any] | tuple[Any] | None = None,
                                   data_as: str = "rows", amount: int | None = None) -> None:
        """
        High-level method overview:
        1) statement preparation + reconnect + execute
        2) queryTypeIn and queryTypeOut
        3) if queryTypeIn isn't empty is means statement was parameterized
        4) collect all information about all columns (such a col types, scales, row_size, etc.) and send it to server
        5) if statement was `select` - perform lists for fetching and do 4 step for queryTypeOut
        """

        if self.__open_statement:
            self.__close_stmt()

        self.__open_statement = True
        self.__more_to_fetch = True

        self.__stmt_id = self.__client.get_statement_id()
        self.__prepare_params = self.__client.prepare_statement(statement)

        if self.__prepare_params.get("reconnect", False):
            # If reconnect exists and issued, close socket, open a new socket with new port/ip sent be the reconnect response
            self.__client.reconnect(self.__stmt_id, self.__conn_params.database, self.__conn_params.service,
                                    self.__conn_params.username, self.__conn_params.password, self.__prepare_params["listener_id"], self.__prepare_params['ip'],
                                    self.__prepare_params['port_ssl'] if self.__conn_params.use_ssl else self.__prepare_params['port'])

        self.__logger.info(f"Executing statement: '{statement}'", connection_id=self.__connection_id, statement_id=self.__stmt_id)
        self.__client.execute_statement()
        self.__column_list = parameterized_columns = self.__client.get_query_type_in()

        if self.__prepare_params.get("local_file_read", False):
            self.__put_file_to_server()

        elif parameterized_columns:
            self.__handle_parameterized(data_as, params, amount)

        self.__query_type_out = self.__client.get_query_type_out()
        self.__column_list = self.__query_type_out.get('queryTypeNamed', [])

        if self.__column_list:  # if there is data in `queryTypeOut`, means it was a `SELECT` query
            self.__statement_type = "SELECT"
            self.__parsed_rows = []
            self.__unparsed_row_amount = 0
            self.__generate_columns_data()
            self.__fill_col_lists()
        else:
            self.__statement_type = "DML"
            self.__close_stmt()

    def __put_file_to_server(self):
        """
        Read local file content and send it to the server in binaries
        """

        local_file = self.__prepare_params["local_file_read"]
        file_limit_mb = self.__prepare_params["staging_area_file_limit_mb"]
        self.__logger.debug(f"Sending local file: '{local_file}' to server")
        self.__logger.debug(f"Checking file size does not exceed allowed file limit mb: {file_limit_mb}")

        try:
            validate_file_size(local_file, file_limit_mb)

            with open(local_file, "rb") as file:
                chunk = file.read(BYTES_PER_FLUSH_LIMIT)

                while chunk:
                    byte_list = [chunk]
                    self.__client.send_data(1, byte_list, len(chunk))
                    chunk = file.read(BYTES_PER_FLUSH_LIMIT)

                self.__client.send_data(0, [], 0)

            self.__logger.debug(f"File '{local_file}' sent successfully.")

        except FileNotFoundError:
            self.__logger.log_and_raise(Error, f"File '{local_file}' not found.")
        except Exception as e:
            self.__logger.log_and_raise(Error, f"Error sending file '{local_file}': {e}")

    def __get_file_from_server(self):
        """
        Get binaries from server and write them to a local file
        If the file already exists, an incremental number is added to the filename.
        """

        original_file_to_write = self.__prepare_params["local_file_write"]
        file_to_write = get_unique_filename(original_file_to_write)

        file_size = self.__query_type_out.get("getFileSize", 0)
        self.__logger.debug(f"Reading file from server and writing to local path: {file_to_write}")

        try:
            validate_disk_space(file_to_write, file_size)

            with open(file_to_write, 'wb') as output_stream:
                self.__fetch()  # fetch fills self.__unsorted_data_columns with a list of bytes we read
                data_buffer = self.__fill_data_buffer()

                while data_buffer and len(data_buffer):
                    output_stream.write(data_buffer)
                    self.__unsorted_data_columns.clear()
                    self.__fetch()
                    data_buffer = self.__fill_data_buffer()

            self.__more_to_fetch = False
            self.__logger.debug(f"Successfully read file from server to path: '{file_to_write}'")

        except Exception as e:
            self.__logger.log_and_raise(Error, f"Failed to save file on local disk: {str(e)}")

    def __fill_data_buffer(self):
        return self.__unsorted_data_columns[0] if len(self.__unsorted_data_columns) > 0 else None

    def __handle_parameterized(self, data_as: str, params: list[Any] | tuple[Any], amount: int | None) -> None:
        """
        prepares the data and send it to the server
        """

        # Check if arrays are allowed before executing the rest
        if not self.__validate_arrays_usage():
            self.__logger.log_and_raise(ArraysAreDisabled, "Arrays are disabled in this connection.",
                                        connection_id=self.__connection_id, statement_id=self.__stmt_id)

        self.__generate_columns_data()
        self.__fill_col_lists()

        self.__row_size = sum([sum(self.__col_sizes),
                               sum(null for null in self.__col_nul if null is True),
                               sum(tvc for tvc in self.__col_tvc if tvc is True)
                               ])

        if self.__row_size * ROWS_PER_FLUSH <= BYTES_PER_FLUSH_LIMIT:
            self.__rows_per_flush = int(ROWS_PER_FLUSH)
        else:
            self.__rows_per_flush = int(BYTES_PER_FLUSH_LIMIT / self.__row_size)

        self.__buffer.clear()

        if data_as == 'alchemy_flat_list':
            # Unflatten SQLalchemy data list
            row_len = len(self.__column_list)
            rows_or_cols = [params[i: i + row_len] for i in range(0, len(params), row_len)]
            data_as = 'rows'
        else:
            rows_or_cols = params

        if 'numpy' in repr(type(params[0])):
            data_as = 'numpy'

        column_lengths = [len(row_or_col) for row_or_col in rows_or_cols]

        if column_lengths.count(column_lengths[0]) != len(column_lengths):
            self.__logger.log_and_raise(ProgrammingError, "Inconsistent data sequences passed for inserting. Please use rows/columns of consistent length",
                                        connection_id=self.__connection_id, statement_id=self.__stmt_id)

        if data_as == 'rows':
            self.__capacity = amount or len(rows_or_cols)
            self.__cols = list(zip(*rows_or_cols))
        else:
            self.__cols = rows_or_cols
            self.__capacity = len(self.__cols)

        self.__send_columns()
        self.__column_list.clear()

    def __fill_col_lists(self) -> None:
        """
        fills the column list information - types, sizes and scales
        """

        self.__col_types.clear()
        self.__col_sizes.clear()
        self.__col_scales.clear()

        for type_tup in self.__col_type_tups:
            is_array = 'ftArray' in type_tup
            offset = 0
            _type = type_tup[0]
            if is_array:
                # for array other stuff like scale is shifted in type_tup
                offset = 1
                # Use tuple for ftArray that will be checked
                # only in buffer like 'ftArray' in col_type
                _type = type_tup[0:2]
            self.__col_types.append(_type)
            self.__col_sizes.append(type_tup[1 + offset] if type_tup[1 + offset] != 0 else TEXT_ITEM_SIZE)
            self.__col_scales.append(type_tup[2 + offset])

    def __validate_arrays_usage(self) -> bool:
        """
        Checks if the executing statement uses arrays and if they are allowed.

        Returns:
            bool: False if arrays are not allowed by connection, but used in
              statement, True otherwise.
        """

        if not self.__allow_array:
            for col in self.__column_list:
                if "ftArray" in col["type"]:
                    return False
        return True

    def __fill_description(self) -> None | List:
        """
        Getting parameters for the cursor's 'description' attribute, even for
        a query that returns no rows. For each column, this includes:
        (name, type_code, display_size, internal_size, precision, scale, null_ok)
        """

        if self.__statement_type != 'SELECT':
            self.description = None
            return self.description

        self.description = []
        for col_name, col_nullalbe, col_type_tup in zip(
                self.__col_names, self.__col_nul, self.__col_type_tups):
            type_code = TYPE_MAPPER.get_typecode(col_type_tup[0])  # Convert SQream type to DBAPI identifier
            display_size = internal_size = col_type_tup[
                1]  # Check if other size is available from API
            precision = 38
            scale = col_type_tup[2]

            self.description.append(
                (col_name, type_code, display_size, internal_size, precision,
                 scale, col_nullalbe))

        return self.description

    def __generate_columns_data(self) -> None:
        """
        In order to send data to server in parameterized queries or when we want to fetch data,
        we need to fill all attributes in the list below:
        1) col_names - list of column names
        2) col_tvc - list of `isTrueVarChar` values
        3) col_nul - list of nullable column properties
        4) col_type_tups - list with column types description (type, size, scale)
        5) col_names_map - dictionary with col_name and index (from 0) for every column

        Content could be received from 'queryTypeIn' or 'queryTypeOut' requests
        Server returns a list of columns data that looks like this:
        { "isTrueVarChar": false, "nullable": true, "type": ["ftInt", 4, 0] }
        """

        self.__col_names = [col.get("name", "") for col in self.__column_list]
        self.__col_tvc = [col["isTrueVarChar"] for col in self.__column_list]
        self.__col_nul = [col["nullable"] for col in self.__column_list]
        self.__col_type_tups = [col["type"] for col in self.__column_list]
        self.__col_names_map = {name: idx for idx, name in enumerate(self.__col_names)}

    def __send_columns(self) -> None:
        """
        Used for parameterized statements.
        After information about all columns was collected by `_execute_sqream_statement` and
        `_generate_columns_data_for_parameterized_statement` methods

        We need to send these data to server to make it inserts parameters via
        slicing a chunk of columns and pass to _send_column_chunk()
        """

        start_idx = 0
        while self.__cols != [()]:
            col_chunk = [col[start_idx:start_idx + self.__rows_per_flush] for col in self.__cols]
            chunk_len = len(col_chunk[0])
            if chunk_len == 0:
                break
            self.__send_column_chunk(col_chunk, chunk_len)
            start_idx += self.__rows_per_flush
            del col_chunk

            self.__logger.debug(f"Sent {chunk_len} rows of data", connection_id=self.__connection_id, statement_id=self.__stmt_id)

    def __send_column_chunk(self, cols: List = None, capacity: int = None) -> None:
        """
        Perform parameterized query - "put" json, header, binarized columns. Used by executemany()
        """

        cols = cols or self.__cols
        cols = cols if isinstance(cols, (list, tuple, set, dict)) else list(cols)

        capacity = capacity or self.__capacity
        # Send columns and metadata to be packed into our buffer
        packed_cols = self.__buffer.pack_columns(cols, capacity, self.__col_types,
                                                 self.__col_sizes, self.__col_nul,
                                                 self.__col_tvc, self.__col_scales)
        byte_count = functools.reduce(lambda c, n: c + len(n), packed_cols, 0)
        self.__client.send_data(capacity, packed_cols, byte_count)

    def __fetch(self):
        """
        sends fetch request to the server, keeps the unsorted data columns list
        """

        fetch_meta = self.__client.fetch()
        num_rows_fetched, column_sizes = fetch_meta['rows'], fetch_meta['colSzs']

        if num_rows_fetched == 0:
            self.__close_stmt()
            return num_rows_fetched

        # Get preceding header
        self.__client.receive(10)

        # Get data as memoryviews of bytearrays.
        self.__unsorted_data_columns = [memoryview(self.__client.receive(size)) for idx, size in enumerate(column_sizes)]
        self.__unparsed_row_amount = num_rows_fetched

    def __fetch_and_parse(self, requested_row_amount: int, data_as: str = "rows") -> None:
        """
        See if this amount of data is available or a fetch from sqream is required
        -1 - fetch all available data. Used by fetchmany()
        """

        if data_as == 'rows':
            self.__buffer.clear()

            self.__fetch()
            self.__more_to_fetch = bool(self.__unparsed_row_amount)  # __fetch() updates self.unparsed_row_amount

            while (requested_row_amount > len(self.__parsed_rows) or requested_row_amount == -1) and self.__more_to_fetch:
                unpacked_columns = self.__buffer.unpack_columns(
                    unsorted_data_columns=self.__unsorted_data_columns,
                    col_types=self.__col_types,
                    col_sizes=self.__col_sizes,
                    col_nul=self.__col_nul,
                    col_tvc=self.__col_tvc,
                    col_scales=self.__col_scales
                )

                self.__parsed_rows.extend(zip(*unpacked_columns))
                self.__unparsed_row_amount = 0
                self.__fetch()
                self.__more_to_fetch = bool(self.__unparsed_row_amount)  # _fetch() updates self.unparsed_row_amount

            self.__buffer.clear()
            self.__unsorted_data_columns.clear()

    def __close_stmt(self) -> None:
        """
        Closes open statement with SQREAM

        Raises:
            ProgrammingError: If server responds with invalid JSON
            ProgrammingError: If server responds with valid JSON,
              but it is not object
            OperationalError: If server responds with "error" key in JSON
        """

        if self.__open_statement:
            self.__logger.info("Closing statement", connection_id=self.__connection_id, statement_id=self.__stmt_id)
            self.__client.close_statement()
            self.__open_statement = False
            self.__buffer.clear()
            self.__logger.debug("Done executing statement", connection_id=self.__connection_id, statement_id=self.__stmt_id)



"""Test arrays disabled by default and raises on fetch and network insert"""
import pytest
from pysqream import connect

from tests.utils import ALL_TYPES, SIMPLE_VALUES, select


TEMP_TABLE = "test_arrays_allowed_temp"


@pytest.mark.parametrize("data_type", ALL_TYPES)
def test_array_enabled_false_default_works_on_select(sqream_cursor, data_type):
    """Test select doesn't fail from array columns with allow_array=False"""
    # Do not use ensure_empty_table because it calls select, but create
    # should not fall
    sqream_cursor.execute(f"CREATE OR REPLACE TABLE {TEMP_TABLE} (x {data_type}[])")
    assert select(sqream_cursor, TEMP_TABLE) == []


@pytest.mark.parametrize("data_type, data", zip(ALL_TYPES, SIMPLE_VALUES))
def test_array_enabled_false_default_works_on_network_insert(sqream_cursor, data_type, data):
    """Test network insert doesn't fail from array columns with allow_array=False"""
    sqream_cursor.execute(f"CREATE OR REPLACE TABLE {TEMP_TABLE} (x {data_type}[])")
    sqream_cursor.executemany(f"insert into {TEMP_TABLE} values (?)", [(data,)])

    assert select(sqream_cursor, TEMP_TABLE) == [(data,)]


@pytest.mark.parametrize('flag', [False, True])
def test_cursor_connection_get_the_same_allow_array(ip, port, flag, conn):
    """Test that allow_array flag is passed to new connection of cursor"""
    # ip is a legacy name that doesn't conform naming style so:
    # pylint: disable=invalid-name
    conn = connect(ip, port, 'master', 'sqream', 'sqream', log=True, allow_array=flag)
    assert conn._Connection__allow_array is flag

    cur = conn.cursor()
    assert cur._Cursor__allow_array is flag
    # cleanup
    cur.close()
    conn.close()



import os
import pytest
from unittest.mock import patch
from pysqream.utils import validate_file_size


class TestFileValidation:
    """
    Tests for validate_file_size function
    """

    def test_file_under_limit(self, tmp_path):
        """
        Test that a file under the size limit passes validation
        """

        test_file = tmp_path / "small_file.txt"
        test_file.write_text("This is a small test file")

        validate_file_size(str(test_file), 1)  # 1MB limit, should not raise any exception

    def test_file_over_limit(self, tmp_path):
        """
        Test that a file over the size limit raises RuntimeError
        """

        test_file = tmp_path / "big_file.bin"

        # Write just enough data to exceed the small limit (0.01MB = ~10KB)
        with open(test_file, 'wb') as f:
            f.write(b'0' * 15_000)  # writing 15KB of data

        with pytest.raises(RuntimeError) as excinfo:
            validate_file_size(str(test_file), 0.01)  # 0.01MB limit

        assert "File size exceeds the allowed limit of 0.01 MB" in str(excinfo.value)

    def test_file_exactly_at_limit(self, tmp_path):
        """
        Test a file exactly at the size limit passes validation
        """

        test_file = tmp_path / "exact_file.bin"

        with open(test_file, 'wb') as f:
            f.write(b'0' * 104_858)

        precise_limit = os.path.getsize(str(test_file))

        validate_file_size(str(test_file), precise_limit)

    def test_nonexistent_file(self):
        """
        Test that a non-existent file raises FileNotFoundError
        """

        non_existent_file = "/path/to/nonexistent/file.txt"

        with pytest.raises(FileNotFoundError):
            validate_file_size(non_existent_file, 10)

    def test_zero_size_file(self, tmp_path):
        """
        Test validation with an empty file. should pass with any limit.
        """

        empty_file = tmp_path / "empty.txt"
        empty_file.touch()

        validate_file_size(str(empty_file), 0.1)
        validate_file_size(str(empty_file), 0)

    def test_zero_limit(self, tmp_path):
        """
        Test validation with zero limit
        """

        test_file = tmp_path / "minimal.txt"
        test_file.write_text("Just a few bytes")

        with pytest.raises(RuntimeError) as excinfo:
            validate_file_size(str(test_file), 0)

        assert "File size exceeds the allowed limit of 0 MB" in str(excinfo.value)

    @patch('os.path.getsize')
    def test_large_file_mocked(self, mock_getsize):
        """
        Test with a mocked very large file (avoiding actual large file creation)
        """

        # Mock a 100 MB file
        mock_getsize.return_value = 104_857_600  # 100 MB in bytes

        # Test with 50 MB limit (should fail)
        with pytest.raises(RuntimeError) as excinfo:
            validate_file_size("mock_large_file.bin", 50)

        assert "File size exceeds the allowed limit of 50 MB" in str(excinfo.value)

        # Test with 200 MB limit (should pass)
        validate_file_size("mock_large_file.bin", 200)

    def test_negative_limit(self, tmp_path):
        """
        Test validation with negative limit (should raise ValueError)
        """

        test_file = tmp_path / "test.txt"
        test_file.write_text("Some content")

        # A negative limit doesn't make logical sense, but we should test how the function handles it
        # Expect it to treat negative limits as "exceeded" since any file size is greater than a negative number
        with pytest.raises(RuntimeError):
            validate_file_size(str(test_file), -1)



from setuptools import setup, find_packages

with open("README.rst", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup_params = dict(
    name='pysqream',
    version='6.2.0',
    description='DB-API connector for SQream DB',
    long_description=long_description,
    url="https://github.com/SQream/pysqream",
    author="SQream",
    author_email="info@sqream.com",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    keywords='database db-api sqream sqreamdb',
    python_requires='>=3.9',
    install_requires=["numpy==1.26.4", "packaging>=24.1", "pyarrow==15.0.0",
                      "pandas==2.2.2", "Faker>=28.0.0", "numexpr==2.8.5"]
)

if __name__ == '__main__':
    setup(**setup_params)



from struct import pack_into
from typing import Any, List
from pysqream.globals import TYPE_MAPPER
from pysqream.column_packers.column_packer import ColumnPacker, ColumnData


class NumberPacker(ColumnPacker):
    """
    Packs all numbers data types except for decimal(numeric)
    """

    def _pack_data(self, data: List[Any]) -> bytes:
        pack_into(f'{len(data)}{TYPE_MAPPER.get_pack_code(self.metadata.col_type)}',
                  memoryview(self.buf_map), self.buf_idx, *data)

        return self.buf_map[:self.buf_idx + len(data) * self.metadata.size]

    def unpack(self, column_data: ColumnData) -> List[Any]:
        self._cast_data_by_pack_code(column_data)
        unpacked_data = [None if self._is_null(n) else d for d, n in zip(column_data.data_column, column_data.nullable)] if self.metadata.nullable \
            else column_data.data_column

        if self.metadata.nullable:
            null_mask = [self._is_null(x) for x in column_data.nullable]
            return [None if is_null else value
                    for is_null, value in zip(null_mask, unpacked_data)]

        return unpacked_data



import pytest
import struct
import pandas as pd
from pysqream import casting
from datetime import datetime
from tests.test_base import TestBaseWithoutBeforeAfter


class TestDatetimeUnitTest(TestBaseWithoutBeforeAfter):

    def test_zero_date(self):
        with pytest.raises(Exception, match="year 0 is out of range"):
            casting.sq_date_to_py_date(0, is_null=False)

    def test_zero_datetime(self):
        with pytest.raises(Exception, match="year 0 is out of range"):
            casting.sq_datetime_to_py_datetime(0, is_null=False)

    def test_negative_date(self):
        with pytest.raises(Exception, match="year -9 is out of range"):
            casting.sq_date_to_py_date(-3000, is_null=False)

    def test_negative_datetime(self):
        with pytest.raises(Exception, match="year 0 is out of range"):
            casting.sq_datetime_to_py_datetime(-3000, is_null=False)

    def test_null_date(self):
        res = casting.sq_date_to_py_date(-3000, is_null=True)
        assert res is None, f"Excepted to get None, but got [{res}]"

    def test_null_datetime(self):
        res = casting.sq_datetime_to_py_datetime(-3000, is_null=True)
        assert res is None, f"Excepted to get None, but got [{res}]"

    insert_data = [
        datetime(2015, 12, 24, 13, 11, 23, 0),
        datetime(2015, 12, 31, 23, 59, 59, 999),
        datetime(2015, 12, 24, 13, 11, 23, 1000),
        datetime(2015, 12, 24, 13, 11, 23, 1500),
        datetime(2015, 12, 31, 23, 59, 59, 2000),
    ]
    expected_result = [
        datetime(2015, 12, 24, 13, 11, 23),
        datetime(2015, 12, 31, 23, 59, 59, 1000),
        datetime(2015, 12, 24, 13, 11, 23, 1000),
        datetime(2015, 12, 24, 13, 11, 23, 2000),
        datetime(2015, 12, 31, 23, 59, 59, 2000),
    ]

    @pytest.mark.parametrize("input_datetime, expected_result", zip(insert_data, expected_result))
    def test_datetime_milliseconds_round(self, input_datetime, expected_result):
        """test case for SQ-13969"""

        dt_as_long = casting.datetime_to_long(input_datetime)
        long_as_dt = casting.sq_datetime_to_py_datetime(dt_as_long)

        assert long_as_dt == expected_result


class TestDatetime2UnitTest(TestBaseWithoutBeforeAfter):

    def test_pandas_timestamp_to_sq_datetime2(self):
        # Test regular datetime conversion
        dt = pd.Timestamp("1997-12-31 23:59:59.123456789")
        utc_offset, nanos, time_int, date_int = casting.pandas_timestamp_to_sq_datetime2(dt)

        assert utc_offset == 0
        assert nanos == 456789  # Only sub-millisecond part
        assert time_int == 86399123  # 23:59:59.123
        assert date_int == 729694  # 1997-12-31

        # Test None input
        result = casting.pandas_timestamp_to_sq_datetime2(None)
        assert result == casting.SQREAM_OLDEST_DATETIME2

        # Test midnight
        dt = pd.Timestamp("2000-01-01 00:00:00")
        utc_offset, nanos, time_int, date_int = casting.pandas_timestamp_to_sq_datetime2(dt)
        assert time_int == 0
        assert nanos == 0

    def test_sq_datetime2_to_pandas_timestamp(self):
        """Test that converts bytes coming from SQream to pandas timestamp and back to bytes"""
        example_bytes = b'\x00\x00\x00\x00U\xf8\x06\x00\x93X&\x05^"\x0b\x00'
        result = casting.sq_datetime2_to_pandas_timestamp(example_bytes, False)

        components = casting.pandas_timestamp_to_sq_datetime2(result)
        datetime2_bytes = struct.pack('<iiii', *components)
        assert datetime2_bytes == example_bytes

        # Test null case
        assert casting.sq_datetime2_to_pandas_timestamp(example_bytes, True) is None

    def test_roundtrip_conversion(self):
        """Test that converting to SQream format and back gives the same datetime"""
        original_dt = pd.Timestamp("1997-12-31 23:59:59.123456789")

        components = casting.pandas_timestamp_to_sq_datetime2(original_dt)
        datetime2_bytes = struct.pack('<iiii', *components)

        result_dt = casting.sq_datetime2_to_pandas_timestamp(datetime2_bytes, False)

        assert str(result_dt) == str(original_dt) + "+00:00"

    @pytest.mark.parametrize("dt", [
        pd.Timestamp("1970-01-17 18:56:04.205739Z"),
        pd.Timestamp("1970-01-17 18:56:04.205739789+00:00"),
        pd.Timestamp("1970-01-17 18:56:04.205739789+02:00"),
        pd.Timestamp("1970-01-17 18:56:04.205739789 +05:30"),
        pd.Timestamp("1970-01-17 18:56:04.205739789-05:00"),
        pd.Timestamp("1970-01-17 18:56:04.205739789 -08:00"),
    ])
    def test_datetime2_timezone_conversions(self, dt):
        """
        checking UTC time, Positive time zone offsets and Negative time zone offsets
        Verifies timezone information is preserved through conversion
        """

        components = casting.pandas_timestamp_to_sq_datetime2(dt)
        datetime2_bytes = struct.pack('<iiii', *components)
        result_dt = casting.sq_datetime2_to_pandas_timestamp(datetime2_bytes, False)

        assert str(result_dt) == str(dt)



import pytest
from unittest.mock import Mock
from struct import pack
from pysqream.globals import PROTOCOL_VERSION, SUPPORTED_PROTOCOLS
from tests.unit_tests.mocks import SocketMock, SQClients


class TestClient:
    def test_receive_full_data(self):
        """
        Test receiving complete data from socket
        """

        test_data = b'Hello World'
        socket_mock = SocketMock(test_data)
        client = SQClients(socket_mock)
        result = client.receive(len(test_data))

        assert result == test_data

    def test_receive_interrupted_connection(self):
        """
        Test handling of interrupted connection (receive returns 0)
        """

        socket_mock = SocketMock(b'')
        client = SQClients(socket_mock)

        with pytest.raises(ConnectionRefusedError) as exc_info:
            client.receive(10)

        assert 'SQreamd connection interrupted' in str(exc_info.value)

    def test_get_text_response(self):
        """
        Test getting text response with valid protocol
        """

        message = "test message"
        message_bytes = message.encode('utf8')
        header = pack('bb', PROTOCOL_VERSION, 1) + pack('q', len(message_bytes))
        socket_mock = SocketMock(header + message_bytes)
        client = SQClients(socket_mock)
        response = client.get_response()

        assert response == message

    def test_protocol_mismatch(self):
        """
        Test handling of unsupported protocol version
        """

        invalid_protocol = max(SUPPORTED_PROTOCOLS) + 1
        header = pack('bb', invalid_protocol, 1) + pack('q', 10)
        socket_mock = SocketMock(header + b'0123456789')
        client = SQClients(socket_mock)

        with pytest.raises(Exception) as exc_info:
            client.get_response()

        assert 'Protocol mismatch' in str(exc_info.value)

    def test_send_string_basic(self):
        """
        Test basic string sending functionality
        """

        socket_mock = SocketMock(b'{"response": "ok"}')
        client = SQClients(socket_mock)
        test_json = '{"test": "value"}'
        client.send_string(test_json, get_response=False)
        sent_data = b''.join(socket_mock.sent_data)

        assert len(sent_data) == 10 + len(test_json)  # header + message
        assert test_json.encode('utf8') in sent_data

    def test_send_string_with_response(self):
        """
        Test sending string and getting response
        """

        response = '{"response": "ok"}'
        response_bytes = response.encode('utf8')
        header = pack('bb', PROTOCOL_VERSION, 1) + pack('q', len(response_bytes))
        socket_mock = SocketMock(header + response_bytes)
        client = SQClients(socket_mock)
        result = client.send_string('{"test": "value"}')

        assert result == response

    def test_prepare_statement_success(self):
        """
        Test successful statement preparation
        """

        response = '{"statementPrepared": true, "chunkSize": 1048576, "canSupportParams": true}'
        response_bytes = response.encode('utf8')
        header = pack('bb', PROTOCOL_VERSION, 1) + pack('q', len(response_bytes))
        socket_mock = SocketMock(header + response_bytes)
        client = SQClients(socket_mock)
        result = client.prepare_statement("SELECT * FROM table")

        assert result["statementPrepared"] is True
        assert result["canSupportParams"] is True
        assert result["chunkSize"] == 1048576

    def test_prepare_statement_error(self):
        """
        Test handling of preparation error
        """

        response = '{"error": "Invalid SQL"}'
        response_bytes = response.encode('utf8')
        header = pack('bb', PROTOCOL_VERSION, 1) + pack('q', len(response_bytes))
        socket_mock = SocketMock(header + response_bytes)
        client = SQClients(socket_mock)

        with pytest.raises(Exception) as exc_info:
            client.prepare_statement("INVALID SQL")

        assert 'Invalid SQL' in str(exc_info.value)

    def test_execute_success(self):
        """
        Test successful statement execution
        """

        response = '{"executed": "executed"}'
        response_bytes = response.encode('utf8')
        header = pack('bb', PROTOCOL_VERSION, 1) + pack('q', len(response_bytes))
        socket_mock = SocketMock(header + response_bytes)
        client = SQClients(socket_mock)
        client.execute_statement()
        sent_data = b''.join(socket_mock.sent_data)

        assert b'{"execute" : "execute"}' in sent_data

    def test_execute_error(self):
        """
        Test handling of execution error
        """

        response = '{"error": "Execution failed"}'
        response_bytes = response.encode('utf8')
        header = pack('bb', PROTOCOL_VERSION, 1) + pack('q', len(response_bytes))
        socket_mock = SocketMock(header + response_bytes)
        client = SQClients(socket_mock)

        with pytest.raises(Exception) as exc_info:
            client.execute_statement()

        assert 'Execution failed' in str(exc_info.value)

    def test_reconnect_success(self):
        """
        Test successful reconnection
        """

        reconnect_response = '{"reconnected": true}'
        reconstruct_response = '{"statementReconstructed": true}'

        responses = []
        for resp in [reconnect_response, reconstruct_response]:
            resp_bytes = resp.encode('utf8')
            header = pack('bb', PROTOCOL_VERSION, 1) + pack('q', len(resp_bytes))
            responses.extend([header, resp_bytes])

        socket_mock = SocketMock(b''.join(responses))
        client = SQClients(socket_mock)

        client.reconnect(
            statement_id=1,
            database="test_db",
            service="test_service",
            username="test_user",
            password="test_pass",
            listener_id=3,
            ip="127.0.0.1",
            port=5000
        )

        sent_data = b''.join(socket_mock.sent_data)
        assert b'"reconnectDatabase"' in sent_data
        assert b'"reconstructStatement"' in sent_data

    def test_reconnect_gpu_error(self):
        """
        Test handling of GPU requirement error during reconnect
        """

        error_msg = 'The query requires a GPU-Worker. Ensure the SQream Service has GPU'

        socket_mock = SocketMock()
        client = SQClients(socket_mock)

        client.send_string = Mock(return_value=f'{{"error": "{error_msg}"}}')

        with pytest.raises(Exception) as exc_info:
            client.reconnect(1, "test_db", "test_service", "user", "pass", 3, "127.0.0.1", 5000)

        assert error_msg in str(exc_info.value)



import threading
from queue import Queue
from tests.test_base import TestBaseWithoutBeforeAfter, connect_dbapi, Logger

logger = Logger()
q = Queue()


class TestThreads(TestBaseWithoutBeforeAfter):
    def connect_and_execute(self, num, con):
        cur = con.cursor()
        cur.execute("select {}".format(num))
        res = cur.fetchall()
        q.put(res)
        cur.close()

    def test_threads(self):
        con = connect_dbapi(self.ip, clustered=self.clustered, use_ssl=self.use_ssl, port=self.port,
                            picker_port=self.picker_port, database=self.database,
                            username=self.username, password=self.password)

        logger.info("Thread tests - concurrent inserts with multiple threads through cursor")
        t1 = threading.Thread(target=self.connect_and_execute, args=(3, con,))
        t2 = threading.Thread(target=self.connect_and_execute, args=(3, con,))
        t1.start()
        t2.start()
        t1.join()
        t2.join()
        res1 = q.get()[0][0]
        res2 = q.get()[0][0]
        con.close()

        assert res1 == res2, f"expected to get equal values. instead got res1: {res1} and res2: {res2}"



from __future__ import annotations

import time
import socket
from typing import Union, List, Dict
from struct import unpack
from pysqream.cursor import Cursor
from pysqream.server.sqsocket import SQSocket
from pysqream.server.sqclient import SQClient
from pysqream.server.connection_params import ConnectionParams
from pysqream.utils import ProgrammingError, Error, dbapi_method
from pysqream.globals import DEFAULT_LOG_PATH
from pysqream.logger import ContextLogger


class Connection:
    """
    Connection class used to interact with SQreamDB.
    The very first connection is called the "base connection".
    For every cursor we open, we create another connection since in sqream every statement should run in its own connection
    This another connection is called "sub connection".
    Every sub connection holds its cursor in cursors dict.
    The base connection holds all sub connections in sub_connections list.
    """

    def __init__(self, conn_params: ConnectionParams, log: Union[bool, str] = False, base_connection: bool = True,
                 reconnect_attempts: int = 3, reconnect_interval: int = 10, allow_array: bool = True, logging_level: str = "INFO",
                 context_logger: Union[ContextLogger] = None):
        self.__version: Union[str, None] = None
        self.__is_connection_closed: bool = False
        self.__connection_id: Union[int, None] = None
        self.__connect_to_socket: bool = False
        self.__connect_to_database: bool = False
        self.__reconnect_attempts: int = reconnect_attempts
        self.__reconnect_interval: int = reconnect_interval
        self.__base_connection: bool = base_connection
        self.__client: Union[SQClient, None] = None
        self.__cursors: Dict[int, Cursor] = {}
        self.__sub_connections: List[Connection] = []

        # SQreamDB connection parameters attributes
        self.__conn_params: ConnectionParams = conn_params
        self.__allow_array: bool = allow_array

        self.__logger = context_logger if context_logger else self.__set_logger(log, logging_level)
        self.__validate_attributes()
        self.__open_connection()
        self.__connect_database()

    def __set_logger(self, log: Union[bool, str], logging_level: str) -> ContextLogger:
        self.__logger = ContextLogger()

        if log:
            self.__logger.start_logging(DEFAULT_LOG_PATH if log is True else log, logging_level=logging_level)

        return self.__logger

    def __enter__(self) -> Connection:
        """
        Implementation for context manager ("with" clause)
        """

        return self

    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:
        """
        Implementation for context manager ("with" clause)
        """

        self.close()

    def __del__(self) -> None:
        """
        Finalizer for connection object. Closes all sub connections and their cursors
        """

        if not self.__is_connection_closed:
            try:
                self.__logger.debug("Try to destroy open connections", connection_id=self.__connection_id)
                self.close()
            except Exception as e:
                if "Trying to close a connection that's already closed" not in repr(e):
                    self.__logger.log_and_raise(ProgrammingError, e, connection_id=self.__connection_id)

        if self.__base_connection:
            self.__logger.stop_logging()

    def __iter__(self) -> Connection:
        """
        Implementation for iterating connection in for-in clause
        """

        for sub_conn in self.__sub_connections:
            yield sub_conn

    # DB-API Must have methods
    # ------------------------
    @dbapi_method
    def close(self) -> None:
        """
        If we are in base connection - iterate every sub connection and call its close method
        then, for every cursor we have in our connection, we call cursor close method
        """

        if not self.__connect_to_database:
            if self.__connect_to_socket:
                return self.__client.disconnect_socket()
            return

        if self.__base_connection:
            for sub_conn in self.__sub_connections:
                sub_conn.close()

        for con_id, cursor in self.__cursors.items():
            try:
                if not cursor.get_is_cursor_closed():
                    cursor._Cursor__is_connection_initiated_close = True
                    cursor.close()
            except Exception as e:
                self.__logger.log_and_raise(Error, f"Can't close connection - {e}", connection_id=con_id)

        self.__cursors.clear()
        self.__sub_connections.clear()
        self.__close_connection()
        self.__is_connection_closed = True

    @dbapi_method
    def cursor(self) -> Cursor:
        """
        Create a new sub-connection with the same connection parameters and create a cursor for that sub connection.
        We use a connection as the equivalent of a 'cursor'
        """

        self.__logger.debug("Create cursor", connection_id=self.__connection_id)
        self.__verify_con_open()
        sub_conn_params = ConnectionParams(self.__conn_params.origin_ip if self.__conn_params.clustered else self.__conn_params.ip,
                                           self.__conn_params.origin_port if self.__conn_params.clustered is True else self.__conn_params.port,
                                           self.__conn_params.database,
                                           self.__conn_params.username,
                                           self.__conn_params.password,
                                           self.__conn_params.clustered,
                                           self.__conn_params.use_ssl,
                                           self.__conn_params.service)
        sub_conn = Connection(
            sub_conn_params,
            base_connection=False,
            reconnect_attempts=self.__reconnect_attempts,
            reconnect_interval=self.__reconnect_interval,
            allow_array=self.__allow_array,
            context_logger=self.__logger
        )
        sub_conn.__verify_con_open()

        cur = Cursor(sub_conn.__conn_params, sub_conn.__client, sub_conn.__connection_id, sub_conn.__allow_array, self.__logger)
        sub_conn.__cursors[cur.get_connection_id()] = cur

        if self.__base_connection:
            self.__sub_connections.append(sub_conn)

        return cur

    @dbapi_method
    def commit(self):
        """
        DB-API requires this method, but SQream doesn't support transactions in the traditional sense
        """

        self.__logger.log_and_raise(NotImplementedError, "Commit called (not supported for SQreamDB)", connection_id=self.__connection_id)

    @dbapi_method
    def rollback(self):
        """
        DB-API requires this method, but SQream doesn't support transactions in the traditional sense
        """

        self.__logger.log_and_raise(NotImplementedError, "Rollback called (not supported for SQreamDB)", connection_id=self.__connection_id)

    # Internal Methods
    # ----------------
    def __validate_attributes(self):
        if not isinstance(self.__reconnect_attempts, int) or self.__reconnect_attempts < 0:
            self.__logger.log_and_raise(Exception, f"reconnect attempts should be a positive integer, got : {self.__reconnect_attempts}")
        if not isinstance(self.__reconnect_interval, int) or self.__reconnect_attempts < 0:
            self.__logger.log_and_raise(Exception, f"reconnect interval should be a positive integer, got : {self.__reconnect_interval}")

    def __open_connection(self) -> None:
        """
        Get proper ip and port from picker if needed and open a socket to the server. Used at __init__()
        If clustered is true -
         - open a non SSL socker for picker communication
         - Read the first 4 bytes to get readlen and read ip, then read 4 more bytes to get the port

        Then create socket and connect to actual SQreamd server
        """

        if self.__conn_params.clustered is True:
            picker_socket = SQSocket(self.__conn_params.origin_ip, self.__conn_params.origin_port, self.__logger, False)
            self.__client = SQClient(picker_socket)
            picker_socket.timeout(5)

            try:
                read_len = unpack('i', self.__client.receive(4))[0]
                picker_socket.timeout(None)
                self.__conn_params.ip = self.__client.receive(read_len)
                self.__conn_params.port = unpack('i', self.__client.receive(4))[0]
                picker_socket.close()
            except socket.timeout:
                self.__logger.log_and_raise(ProgrammingError, f"Connected with clustered=True, but apparently not a server picker port",
                                            connection_id=self.__connection_id)

        sqsocket = SQSocket(self.__conn_params.ip, self.__conn_params.port, self.__logger, self.__conn_params.use_ssl)
        self.__client = SQClient(sqsocket)
        self.__connect_to_socket = True

    def __connect_database(self) -> None:
        """
        Handle connection to database, with or without server picker
        """

        if self.__connect_to_socket:
            try:
                self.__connection_id, self.__version = self.__client.connect_to_socket(self.__conn_params.username,
                                                                                       self.__conn_params.password,
                                                                                       self.__conn_params.database,
                                                                                       self.__conn_params.service)
            except KeyError as e:
                self.__logger.log_and_raise(ProgrammingError, str(e))

            self.__logger.info(f"Connection opened to database {self.__conn_params.database}.", connection_id=self.__connection_id)
            self.__connect_to_database = True

    def __attempt_reconnect(self) -> bool:
        """
        Attempt to reconnect with exponential backoff
        """

        for attempt in range(self.__reconnect_attempts):
            wait_time = self.__reconnect_interval * (2 ** attempt)  # Exponential backoff
            self.__logger.info(f"Waiting {wait_time} seconds before reconnect attempt {attempt + 1}")
            time.sleep(wait_time)

            try:
                self.__logger.info(f"Reconnect attempt {attempt + 1}")
                self.__open_connection()
                self.__connect_database()
                self.__logger.info(f"Reconnection successful on attempt {attempt + 1}")
                return True
            except Exception as e:
                self.__logger.error(f"Reconnect attempt {attempt + 1} failed: {e}")

        self.__logger.log_and_raise(ConnectionRefusedError, f"All {self.__reconnect_attempts} reconnection attempts failed")

    def __close_connection(self) -> None:
        if self.__is_connection_closed:
            self.__logger.log_and_raise(ProgrammingError, f"Trying to close a connection that's already closed for database "
                                                        f"{self.__conn_params.database}", connection_id=self.__connection_id)
        self.__client.close_connection()
        self.__logger.info(f"Connection closed to database {self.__conn_params.database}.", connection_id=self.__connection_id)

    def __verify_con_open(self) -> None:
        if self.__is_connection_closed:
            self.__logger.log_and_raise(ProgrammingError, "Connection has been closed", connection_id=self.__connection_id)
